# Copyright (c) 2019-2023 r3w0p
# The following code can be redistributed and/or
# modified under the terms of the MIT License.

"""
Top-level package for BoboCEP.
"""

__author__ = """r3w0p"""
__email__ = "rr33ww00pp@gmail.com"
__version__ = "1.1.0"

from bobocep.bobocep import BoboError, BoboJSONable, BoboJSONableError
